/*---------------------------------------------------------------------------*/
/*                                                                           */
/* Program name: amex_auth                                                   */
/*                                                                           */
/* Description: AMEX credit card information is passed in. An authorization  */
/*              request transaction is passed to the AMEX server. It         */
/*              retrieves an authorization response transaction from the     */
/*              AMEX server and passes the response back to the caller using */
/*              the value of GOOD or BAD.                                    */
/*                                                                           */
/* Author: Angela Vella & Rich Rogers                                3/14/96 */
/*                                                                           */
/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#include "iso1100.h"
#include "iso1110.h"
#include "ccamexauthserver.h"
#include "ccamexserverutil.h"
#include "transd.h"
#include "order.h"

struct AMEXOrd {
  char                  acct_no[16];
  Iso1100ProcCode       proc_code;
  float                 amount;
  unsigned char         trace[ORDER_NUM_LEN];  /* order number */
  unsigned int          exp_month;
  unsigned int          exp_year;
  Iso1100MerchantType   merchant_type;
  unsigned char         country[4];
  unsigned char         pos_data[13];         /* AMEX provides */
  Iso1100MerchantType   card_acceptor_bus_code;
  unsigned int          approval_code_len;
  unsigned char         card_accept_id[11];   /* AMEX provides */
  struct CardAcceptLoc  acceptor_loc;       /* AMEX provides */
  unsigned char         currency_code[4];
  unsigned char         security_info[8];      /* AMEX provides */
  Iso1110ActionCode     action_code;
  char                  approval_code[APPROVAL_CODE_LEN+1];
};

/*---------------------------------------------------------------------------*/
/* amex_auth                                                                 */
/*---------------------------------------------------------------------------*/
int amex_auth( struct AMEXOrd *amex_order)
{
  Iso1100*               iso1100Ptr;  /* ptr to 1100 object */
  Iso1110*               iso1110Ptr;  /* ptr to 1110 object */
  CcAmexAuthServerPtr    AuthServPtr; /* ptr to AmexAuthServer object */
  CcAmexAuthServerStatus AuthServrc;  /* rc from AmexAuthServer methods */
  Iso1110ActionCode      action_code;
  char                   *byteStreamPtr;
  size_t                 tmOut = 5;   /* wait 5 secs for AMEX response */
  unsigned char          retrievalRef[ISO1100_RETRIEVAL_REF_MAXLEN + 1];
  int                    returnrc;
  char                   errorrec[200];
  char                   hostip[200];
  char                   hostname[200];
  char                   portnum[200];
  char                   debug_ssl[200];
  char                   debug_info[200];
  char                   logfile[200];
  int                    envrc;
  int                    retRef = getpid();
  static int             incr = 1;
  char                   retRefBuf[13];
  time_t                 currTime;
  struct tm*             transTime;

  currTime = time(NULL) ;
  transTime = localtime(&currTime) ;
  sprintf(retRefBuf, "%0*d", 12, retRef+(incr++)) ; /* just to create diff.
	          reference numbers every time */

  /******************************************/
  /* set environment variables              */
  /******************************************/

  strcpy(hostip, "CCAUTH_HOSTIP=");
  strcat(hostip,TCPAUTH_DAEMON_HOSTIP);
  if ((envrc = putenv(hostip)) != 0)
    logger(LOG_ERR, "%s %d %s %s", "Error",envrc,"setting",hostip);

  strcpy(hostname, "CCAUTH_HOSTNAME=");
  strcat(hostname,TCPAUTH_DAEMON_HOSTNAME);
  if ((envrc = putenv(hostname)) != 0)
  logger(LOG_ERR, "%s %d %s %s", "Error",envrc,"setting",hostname);

  strcpy(portnum, "CCAUTH_PORTNUM=");
  strcat(portnum,TCPAUTH_DAEMON_PORTNUM);
  if ((envrc = putenv(portnum)) != 0)
    logger(LOG_ERR, "%s %d %s %s", "Error",envrc,"setting",portnum);

  strcpy(debug_ssl, "CCAUTH_DEBUG_SSL=Yes");
  if ((envrc = putenv(debug_ssl)) != 0)
    logger(LOG_ERR, "%s %d %s %s", "Error",envrc,"setting",debug_ssl);

  strcpy(debug_info, "CCAUTH_DEBUG_INFO=Yes");
  if ((envrc = putenv(debug_info)) != 0)
    logger(LOG_ERR, "%s %d %s %s", "Error",envrc,"setting",debug_info);

  strcpy(logfile, "CCAUTH_LOGFILE=");
  strcat(logfile,"logfile");
  if ((envrc = putenv(logfile)) != 0)
    logger(LOG_ERR, "%s %d %s %s", "Error",envrc,"setting",logfile);

  /* initialize 1100 instance */
  if ((iso1100Ptr  = initIso1100(ISO1100_AMEX)) == NULL)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
      iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
      " -- error initializing ISO1100 object -- Order was NOT accepted.");
    return(BAD);
  }

  /* initialize AmexAuthServ instance */
  if ((AuthServPtr = initCcAmexAuthServer(&AuthServrc)) == 0)
  {
    logger(LOG_ERR, "%s %s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    getAmexErr(AuthServrc),
    " -- error initializing AmexAuthServ object -- Order was NOT accepted.");
    return(BAD);
  }

/******************************************/
/* populate fields in 1100 object         */
/******************************************/
   /*****  account number *****/
  if (iso1100Ptr->setAccountNo(iso1100Ptr, amex_order->acct_no,  ASCII)!=VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting account number -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  processing code *****/
  if (iso1100Ptr->setProcCode(iso1100Ptr, amex_order->proc_code) != VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting process code -- Order was NOT accepted.");
    return(BAD);
  }
 
   /*****  amount of order *****/
  if (iso1100Ptr->setAmount(iso1100Ptr,  amex_order->amount)!=VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
     " -- error setting order amount -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  trace *****/
  if ( iso1100Ptr->setSystemTrace(iso1100Ptr, amex_order->trace,  ASCII) != VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting trace amount -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  effective date hardcoded for now - testing *****/
  if ( iso1100Ptr->setEffective(iso1100Ptr, 94, 8)!=VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting effective date -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  expiration date *****/
  if ( iso1100Ptr->setExpiration(iso1100Ptr, amex_order->exp_year, amex_order->exp_month)!=VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting expiration date -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  merchant type *****/
  if ( iso1100Ptr->setMerchantType(iso1100Ptr, amex_order->merchant_type) != VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting merchant type -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  country code *****/
  if ( iso1100Ptr->setCountry(iso1100Ptr, amex_order->country, ASCII) !=VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting country code -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  Point of Sale (POS) data *****/
  if (iso1100Ptr->setPosDataCode(iso1100Ptr,  amex_order->pos_data,  ASCII)!=VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting POS data -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  card acceptor business code *****/
  if (iso1100Ptr->setCardAccpBusCode(iso1100Ptr,  amex_order->card_acceptor_bus_code)!=VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting business code -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  approval code length  *****/
/*  if (iso1100Ptr->setApprovalCodeLength(iso1100Ptr,  APPROVAL_CODE_LEN)!=VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting approval code -- Order was NOT accepted.");
    return(BAD);
  }*/

   /*****  card acceptor id code *****/
  if (iso1100Ptr->setCardAcceptID(iso1100Ptr, AMEX_SE, amex_order->card_accept_id, ASCII) != VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting id code -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  card acceptor name, location *****/
  if (iso1100Ptr->setCardAcceptLocation(iso1100Ptr, amex_order->acceptor_loc.name, amex_order->acceptor_loc.zip, amex_order->acceptor_loc.region, amex_order->acceptor_loc.country, ASCII) != VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting location -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  currency type *****/
  if (iso1100Ptr->setAmountCurrency(iso1100Ptr, amex_order->currency_code, ASCII)!=VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting currency -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  transaction time *****/
  if (iso1100Ptr->setTransactionTime(iso1100Ptr,  *transTime) != VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting transaction time -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  retrieval reference *****/
  if (iso1100Ptr->setRetrievalRef(iso1100Ptr,  (unsigned char*)retRefBuf, ASCII) != VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting retrieval reference -- Order was NOT accepted.");
    return(BAD);
  }

   /*****  security info commented out until valid info provided *****/
/*  if (iso1100Ptr->setSecurityInfo(iso1100Ptr, amex_order->security_info, ASCII)!=VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error setting security info -- Order was NOT accepted.");
    return(BAD);
  }*/

/*-----------------------------------------------------------------*/
   /*****  validate the data *****/
  if (iso1100Ptr->validate(iso1100Ptr) != VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error validating data -- Order was NOT accepted.");
    return(BAD);
  }

/*  iso1110Ptr = AuthServPtr->authorize(AuthServPtr, iso1100Ptr, tmOut, &AuthServrc);*/
  iso1110Ptr = AuthServPtr->authorize(AuthServPtr, iso1100Ptr, 18, &AuthServrc);

  if (AuthServrc != AMEX_OK)
  {
    logger(LOG_ERR, "%s %s %s %s %d %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    getAmexErr(AuthServrc), " return code", AuthServrc,
    " -- error authorizing -- Order was NOT accepted.");
    return(BAD);
  }

  if ( iso1110Ptr->getActionCode(iso1110Ptr, &(amex_order->action_code)) != VALID)
  {
    logger(LOG_ERR, "%s %s %s", "AMEX",
    iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
    " -- error getting action code -- Order was NOT accepted");
    return(BAD);
  }

  strcpy(amex_order->approval_code," ");

  /* get the authorization code */
  if (amex_order->action_code == APPROVED)
  {
    if (iso1110Ptr->getApprovalCode(iso1110Ptr, ASCII, amex_order->approval_code) != VALID)
    {
      logger(LOG_ERR, "%s %s %s", "AMEX",
      iso1100Ptr->getErrorString(iso1100Ptr, iso1100Ptr->status(iso1100Ptr)),
      " -- error getting approval code -- Order was NOT accepted");
      return(BAD);
    }
  }
  else
  {
    logger(LOG_ERR, "%s %s %s", "Not Approved by AMEX. Action code = ",
    amex_order->action_code, ". Order was NOT accepted.");

    /* get the authorization code */
    if (amex_order->action_code == DENY)
      logger(LOG_ERR, "%s", " Authorization was denied for this purchase amount.");

    /* get the authorization code */
    if (amex_order->action_code == EXPIRED_CARD)
      logger(LOG_ERR, "%s", " Credit Card is expired.");
  }

  /*****  cleanup - delete objects *****/
  deleteCcAmexAuthServer(AuthServPtr);
  deleteIso1100(iso1100Ptr);
  deleteIso1110(iso1110Ptr);

  return(GOOD);
} /* amex_auth */
