#if !defined(__ISO1100_H)
#define __ISO1100_H
/**************************************************************************
 * File Name :		iso1100.h
 * 
 * Description : 	Declaration for class Iso1100
 *
 * Author : 		Ben Lurie
 *
 * Date Created : 	10/29/95
 *
 **************************************************************************
 */

/***************************************************************************
 *                      CLASS    Iso1100
 *
 * DESCRIPTION:
 *
 * The purpose of the Iso1100 class is to provide the ability to create
 * an Iso 8583 Authorization Request (1100) from a 'C' program.
 *
 * The Iso1100 class contains methods to set all of the fields as 
 * specified in the ISO 8583 64-Bit Message Table (from the 
 * American Express Credit Authorization Guide).
 *
 * The Iso1100 class also has a 'validate' method that will validate
 * that:  1) The necessary information for a credit card authorization 
 * has been provided, and, 2) The information is valid (has the correct
 * semantics and syntax).
 *
 * There is only one "get" method:  getRetrievalRef().
 *
 * Use the initIso1100() function to instantiate an Iso1100 object.
 *
 * NOTE:
 *	All methods require passing a pointer to the ojbect to work on (this
 *	is necessary since C does not have real classes.
 *
 * This object is not thread-safe.
 *
 **************************************************************************
 */

#include <time.h>

#include "isomessage.h"

typedef IsoMessageStatus	Iso1100Status;
typedef IsoProcCode			Iso1100ProcCode;
typedef IsoCharacterType	Iso1100CharacterType;
typedef unsigned int		Iso1100MerchantType; 	

typedef enum
{
	AMEX_SE,
	AIRLINE_CODE,
	NCR_NUMBER,
	IATA_AGENT_ID
} Iso1100MerchantIDType;

typedef enum
{
	ISO1100_AMEX,
	ISO1100_VISA,
	ISO1100_MASTERCARD,
	ISO1100_CARD_OTHER
} Iso1100CardType;

typedef struct Iso1100_t
{
	
	/************************************************************/
	/* Public methods.                                          */
	/************************************************************/

	/************************************************************/
	/* Set ISO 1100 data methods.                               */
	/************************************************************/

	/* Bit Position 2 */
	const Iso1100Status	(*setAccountNo)(struct Iso1100_t *, 
										const unsigned char*,
										const Iso1100CharacterType);

	/* Bit Position 4 */
	const Iso1100Status	(*setAmount)(struct Iso1100_t *, const float );

	/* Bit Position 49 
	 * This is the character representation of the ISO 8583 currency
	 * codes.  It can either be the "character" representation (e.g.
	 * USD), or the numeric representation "840".
	 */
	const Iso1100Status	(*setAmountCurrency)(struct Iso1100_t *, 
											 const unsigned char*,
											 const Iso1100CharacterType);

	/* Bit Position 12 */
	const Iso1100Status	(*setTransactionTime)(struct Iso1100_t *, 
											  const struct tm );

	/* Bit Position 13 */
	const Iso1100Status	(*setEffective)(struct Iso1100_t *, 
										const unsigned int year, 
										const unsigned int month);

	/* Bit Position 14 */
	const Iso1100Status	(*setExpiration)(struct Iso1100_t *, 
										 const unsigned int year, 
										 const unsigned int month);

	/* Bit Position 18 */
	const Iso1100Status	(*setMerchantType)(struct Iso1100_t *, 
										   const Iso1100MerchantType );

	/* Bit Position 19 */
	const Iso1100Status	(*setCountry)(struct Iso1100_t *, 
									  const unsigned char* country,
									  const Iso1100CharacterType);

	/* Bit Position 22 */
	const Iso1100Status	(*setPosDataCode)(struct Iso1100_t *, 
										  const unsigned char*,
										  const Iso1100CharacterType);

	/* Bit Position 24 */
	const Iso1100Status	(*setFunctionCode)(struct Iso1100_t *, 
										   const unsigned short );

	/* Bit Position 26 */
	const Iso1100Status	(*setCardAccpBusCode)(struct Iso1100_t *, 
											  const Iso1100MerchantType );

	/* Bit Position 27 */
	const Iso1100Status	(*setApprovalCodeLength)(struct Iso1100_t *, 
												 const unsigned short );

	/* Bit Position 32 */
	const Iso1100Status	(*setAcquireInstID)(struct Iso1100_t *, 
											const unsigned char*,
											const Iso1100CharacterType);

	/* Bit Position 33 */
	const Iso1100Status	(*setForwardInstID)(struct Iso1100_t *, 
											const unsigned char*,
											const Iso1100CharacterType);

	/* Bit Position 45 */
	const Iso1100Status	(*setTrack1Data)(struct Iso1100_t *, 
										 const unsigned char*,
										 const Iso1100CharacterType);

	/* Bit Position 35 */
	const Iso1100Status	(*setTrack2Data)(struct Iso1100_t *, 
										 const unsigned char*,
										 const Iso1100CharacterType);

	/* Bit Position 3 */
	const Iso1100Status	(*setProcCode)(struct Iso1100_t *, 
									   const Iso1100ProcCode );

	/* Bit Position 41 */
	const Iso1100Status	(*setAcceptTermID)(struct Iso1100_t *, 
										   const unsigned char*,
										   const Iso1100CharacterType);

	/* Bit Position 42 */
	const Iso1100Status	(*setCardAcceptID)(struct Iso1100_t *, 
										   const Iso1100MerchantIDType,
										   const unsigned char* id, /* Bit Position 42 */
										   const Iso1100CharacterType); 

	/* Bit Position 43 */
	const Iso1100Status	(*setCardAcceptLocation)(struct Iso1100_t *, 
												 const unsigned char* nameLoc,
												 const unsigned char* zip,
												 const unsigned char* region,
												 const unsigned char* country,
												 const Iso1100CharacterType);

	/* Bit Position 11 */
	const Iso1100Status	(*setSystemTrace)(struct Iso1100_t *, 
										  const unsigned char*,
										  const Iso1100CharacterType);

	/* Bit Position 37 */
	const Iso1100Status	(*setRetrievalRef)(struct Iso1100_t *, 
										   const unsigned char*,
										   const Iso1100CharacterType);

	/* Bit Position 53 */
	const Iso1100Status	(*setSecurityInfo)(struct Iso1100_t *, 
										   const unsigned char*,
										   const Iso1100CharacterType);

	/* Bit Position 62 */
	const Iso1100Status	(*setPrivate1Data)(struct Iso1100_t *, 
										   const unsigned char*,
										   const Iso1100CharacterType);

	/* Bit Position 63 */
	const Iso1100Status	(*setPrivate2Data)(struct Iso1100_t *, 
										   const unsigned char*,
										   const Iso1100CharacterType); 

	/************************************************************/
	/* Get methods.                                           */
	/************************************************************/
	
	/* Bit Position 37 */
	const Iso1100Status (*getRetrievalRef)(struct Iso1100_t*, 
										   Iso1100CharacterType,
										   unsigned char *);

	/************************************************************/
	/* Other methods.                                           */
	/************************************************************/
	const Iso1100Status	(*status)(struct Iso1100_t *);

	/* 
	 * Return a pointer to a byte stream that contains an EBCDIC
	 * representation of an ISO 8583 1100 message.  
	 *
	 * This is what is used by the back-end Amex systems.
	 *
	 * The object must be validated before use (by calling the
	 * validate() method).
	 */
	const void *		(*getByteStream)(struct Iso1100_t *, size_t*);

	/* 
	 * Given an error status code, return a pointer to a string
	 * describing the error.
	 */
	const unsigned char* (*getErrorString)(struct Iso1100_t *, 
										   const Iso1100Status);

	/* Validate the contents of the Iso1100 message. */
	const Iso1100Status (*validate)(struct Iso1100_t *);

	/************************************************************/
	/* Private methods and data.                                */
	/************************************************************/
	/* Bit Position 7 */
	const Iso1100Status	(*setTransmissionTime)(struct Iso1100_t *, 
											   const struct tm );

	Iso1100Status			_status;		/* status of the object */
	void*					byteStreamPtr;

	/*
	 * This is really the internal data of this object.  We don't
	 * define the structure here because that would expose the
	 * internal implementation of this object to users.  Not
	 * only is that bad programming practice, but it would mean
	 * that users would have to recompile every time the data 
	 * changed.
	 *
	 * To avoid those issues, the internal data is defined in
	 * iso1100internal.h
	 */
	void*					dataPtr;

} Iso1100;

/********************************************************/
/* Maximum lengths for string data (not including a NULL
/* character at the end).
/********************************************************/

#define ISO1100_ACCOUNT_NO_BITPOSITION 2
#define ISO1100_ACCOUNT_NO_MAXLEN 19

#define ISO1100_PROCESSING_CODE_BITPOSITION 3
#define ISO1100_PROCESSING_CODE_MAXLEN 6

#define ISO1100_AMOUNT_BITPOSITION 4
#define ISO1100_AMOUNT_MAXLEN 12

#define ISO1100_DATE_TIME_TRANSMISSION_BITPOSITION 7
#define ISO1100_DATE_TIME_TRANSMISSION_MAXLEN 10

#define ISO1100_SYSTEM_TRACE_BITPOSITION 11
#define ISO1100_SYSTEM_TRACE_MAXLEN 6

#define ISO1100_DATE_TIME_TRANSACTION_BITPOSITION 12
#define ISO1100_DATE_TIME_TRANSACTION_MAXLEN 12

#define ISO1100_DATE_EFFECTIVE_BITPOSITION 13
#define ISO1100_DATE_EFFECTIVE_MAXLEN 4

#define ISO1100_DATE_EXPIRATION_BITPOSITION 14
#define ISO1100_DATE_EXPIRATION_MAXLEN 4

#define ISO1100_MERCHANT_TYPE_BITPOSITION 18
#define ISO1100_MERCHANT_TYPE_MAXLEN 4

#define ISO1100_COUNTRY_CODE_BITPOSITION 19
#define ISO1100_COUNTRY_CODE_MAXLEN 3

#define ISO1100_POS_DATA_CODE_BITPOSITION 22
#define ISO1100_POS_DATA_CODE_MAXLEN 12

#define ISO1100_FUNCTION_CODE_BITPOSITION 24
#define ISO1100_FUNCTION_CODE_MAXLEN 3

#define ISO1100_MESSAGE_REASON_CODE_BITPOSITION 25
#define ISO1100_MESSAGE_REASON_CODE_MAXLEN 4

#define ISO1100_CARD_ACCEPTOR_BUS_CODE_BITPOSITION 26
#define ISO1100_CARD_ACCEPTOR_BUS_CODE_MAXLEN 4

#define ISO1100_APPROVAL_CODE_LENGTH_BITPOSITION 27
#define ISO1100_APPROVAL_CODE_LENGTH_MAXLEN 1

#define ISO1100_ACQUIRE_INST_ID_BITPOSITION 32
#define ISO1100_ACQUIRE_INST_ID_MAXLEN 11

#define ISO1100_FORWARD_ID_BITPOSITION 33
#define ISO1100_FORWARD_ID_MAXLEN 11

#define ISO1100_TRACK2_DATA_BITPOSITION 35
#define ISO1100_TRACK2_DATA_MAXLEN 39

#define ISO1100_RETRIEVAL_REF_BITPOSITION 37
#define ISO1100_RETRIEVAL_REF_MAXLEN 12

#define ISO1100_ACCEPT_TERM_ID_BITPOSITION 41
#define ISO1100_ACCEPT_TERM_ID_MAXLEN 8

#define ISO1100_CARD_ACCEPT_ID_BITPOSITION 42
#define ISO1100_CARD_ACCEPT_ID_MAXLEN 15

#define ISO1100_CARD_ACCEPT_LOCATION_NAME_BITPOSITION 43
#define ISO1100_CARD_ACCEPT_LOCATION_NAME_MAXLEN 83

#define ISO1100_CARD_ACCEPT_LOCATION_ZIP_BITPOSITION 43
#define ISO1100_CARD_ACCEPT_LOCATION_ZIP_MAXLEN 10

#define ISO1100_CARD_ACCEPT_LOCATION_REGION_BITPOSITION 43
#define ISO1100_CARD_ACCEPT_LOCATION_REGION_MAXLEN 3

#define ISO1100_CARD_ACCEPT_LOCATION_COUNTRY_BITPOSITION 43
#define ISO1100_CARD_ACCEPT_LOCATION_COUNTRY_MAXLEN 3

#define ISO1100_TRACK1_DATA_BITPOSITION 45
#define ISO1100_TRACK1_DATA_MAXLEN 76

#define ISO1100_CURRENCY_CODE_BITPOSITION 49
#define ISO1100_CURRENCY_CODE_MAXLEN 3

#define ISO1100_SECURITY_INFO_BITPOSITION 53
#define ISO1100_SECURITY_INFO_MAXLEN 8

#define ISO1100_PRIVATE1_DATA_BITPOSITION 62
#define ISO1100_PRIVATE1_DATA_MAXLEN 60

#define ISO1100_PRIVATE2_DATA_BITPOSITION 63
#define ISO1100_PRIVATE2_DATA_MAXLEN 100


/********************************************************/
/* Maximum length of any bitposition.
/********************************************************/
#define ISO1100_BITPOSITION_MAXLEN		(100)

/********************************************************/
/* Instantiate and delete an instance of an ISO 1100 message.
/********************************************************/

extern Iso1100*	initIso1100(const Iso1100CardType);
extern void 	deleteIso1100(Iso1100*);

#endif /* !defined(__ISO1100_H) */

