#if !defined(__ISO1100INTERNAL_H)
#define __ISO1100INTERNAL_H

/**************************************************************************
 * File Name :		iso1100internal.h
 * 
 * Description : 	Internal class definitions for the Iso1100 object
 *
 * Author : 		Ben Lurie
 *
 * Date Created : 	10/29/95
 *
 **************************************************************************
 */

#include "iso1100.h"

/********************************************************/
/* Preprocessor defines 								*/
/********************************************************/

/**************************************************************************
 *
 * NAME:                valueSet
 *
 * ARGUMENTS:           thisPtr
 *							Pointer to Iso1100 message to work on.
 *
 *						bitPosition
 *							Bitposition of data that has been set.
 *
 * RETURN VALUE:		None.
 *
 * SCOPE:               private
 *
 * DESCRIPTION:
 *
 * This macro records the fact that the user has set some piece of
 * data whose bitposition is 'bitPosition'. 
 *
 * The current implementation of this is to set a bit in the 
 * 'valueSetBitmap' field.
 *
 * NOTE:
 *		When a value is set, we must record the fact that this
 *		data must is unvalidated.  This is done by recording
 * 		the fact that the Iso1100 message has not been validated.
 *
 **************************************************************************
 */
#define valueSet(thisPtr, bitPosition) \
	IsoSetBit(((Iso1100Data*)(thisPtr->dataPtr))->valueSetBitmap, \
			  bitPosition); \
	((Iso1100Data*)(thisPtr->dataPtr))->objectValidated = FALSE;

/**************************************************************************
 *
 * NAME:                valueUnset
 *
 * ARGUMENTS:           thisPtr
 *							Pointer to Iso1100 message to work on.
 *
 *						bitPosition
 *							Bitposition of data that has been unset.
 *
 * RETURN VALUE:		None.
 *
 * SCOPE:               private
 *
 * DESCRIPTION:
 *
 * This macro records the fact that the user has unset some piece of
 * data whose bitposition is 'bitPosition'. 
 *
 * The current implementation of this is to unset a bit in the 
 * 'valueSetBitmap' field.
 *
 * NOTE:
 *		When a value is unset, we must record the fact that the
 *		whole Iso1100 object must be revalidated.  This is necessary
 *		because there are dependencies between the data fields and
 *		unsetting some data may cause some other piece of data to
 *		become invalid.
 *
 **************************************************************************
 */
#define valueUnset(thisPtr, bitPosition) \
	IsoClearBit(((Iso1100Data*)(thisPtr->dataPtr))->valueSetBitmap, \
				bitPosition); \
	(((Iso1100Data*)(thisPtr->dataPtr))->objectValidated = FALSE;

/**************************************************************************
 *
 * NAME:                valueIsSet
 *
 * ARGUMENTS:           thisPtr
 *							Pointer to Iso1100 message to work on.
 *
 *						bitPosition
 *							Bitposition of data to be tested.
 *
 * RETURN VALUE:		TRUE if data at 'bitPosition' has been set,
 *						FALSE if data has not been set.
 *
 * SCOPE:               private
 *
 * DESCRIPTION:
 *
 * This macro tests to see if data at a particular bitposition has
 * been set.
 *
 * The current implementation of this is to test a bit in the 
 * 'valueSetBitmap' field.
 *
 **************************************************************************
 */
#define valueIsSet(thisPtr, bitPosition) \
	IsoTestBit(((Iso1100Data*)(thisPtr->dataPtr))->valueSetBitmap, \
			   bitPosition)

/**************************************************************************
 *
 * NAME:                isObjectValid
 *
 * ARGUMENTS:           thisPtr
 *							Pointer to Iso1100 message to work on.
 *
 * RETURN VALUE:		TRUE if object can be used, 
 *						FALSE if object is not valid to use.
 *
 * SCOPE:               private
 *
 * DESCRIPTION:
 *
 * This macro tests to see if the object is valid to use.  If the
 * status is 'INVALID' or if there has been a problem with memory,
 * then the object is not valid to use.
 *
 * This is used by the methods of the Iso1100 object to see if 
 * they are being passed a valid object.
 *
 **************************************************************************
 */
#define isObjectValid(thisPtr) \
	((thisPtr->_status != INVALID) && (thisPtr->_status != ENOMEM))

/*
 * This is the internal data of an Iso1100 message.  
 *
 * It contains one data element for each bit position that can 
 * be set in an Iso1100 message.  The data element types are 
 * typically influenced by how the user is going to pass the information
 * into the 'set' method (and this is influenced by the fact that this
 * information is going to be set programatically from a 'C' program).
 *
 * There is also data elements for internal use as well.
 *
 * NOTE:
 * 		The main purpose of having the data elements reside in the
 *		Iso1100 message is for:
 *			
 *			*	efficient retrieval.  We don't have to get the information
 *				from the Iso byte stream object and reformat it for
 *				the user.
 *
 *			*	Efficient setting to the IsoMessageByteStream object.
 *				The IsoMessageByteStream object works most efficiently
 *				if data elements are accessed or set in ascending
 *				bit position order.  Since it is unlikely that the
 *				user of an Iso1100 object is going to set the data
 *				in ascending bit position order, we save the data
 *				internally and only set it when necessary, and then
 *				in an efficient manner.
 *
 * 		This is basically a time versus space tradeoff.
 */
typedef struct Iso1100Data_t
{
	unsigned char	accountNo[ISO1100_ACCOUNT_NO_MAXLEN + 1];	/* BP 2 */
	float			amount;										/* BP 4 */
	unsigned char	currencyCode[ISO1100_CURRENCY_CODE_MAXLEN + 1];/* BP 49 */
	struct tm		transactionTime;							/* BP 12 */
	struct tm		transmissionTime;							/* BP 7 */
	unsigned int	effectiveYear;								/* BP 13 */
	unsigned int	effectiveMonth;								/* BP 13 */
	unsigned int	expirationYear;								/* BP 14 */
	unsigned int	expirationMonth;							/* BP 14 */
	Iso1100MerchantType	merchantType;							/* BP 18 */
	unsigned char	country[ISO1100_COUNTRY_CODE_MAXLEN + 1];	/* BP 19 */
	unsigned char	posDataCode[ISO1100_POS_DATA_CODE_MAXLEN + 1]; /* BP 22 */
	unsigned short	functionCode;								/* BP 24 */
	Iso1100MerchantType	cardAccpBusCode;						/* BP 26 */
	unsigned short	approvalCodeLength;							/* BP 27 */
	unsigned char	acquireInstId[ISO1100_ACQUIRE_INST_ID_MAXLEN+1];/* BP 32 */
	unsigned char	forwardInstId[ISO1100_FORWARD_ID_MAXLEN + 1];	/* BP 33 */
	unsigned char	track1Data[ISO1100_TRACK1_DATA_MAXLEN + 1];	/* BP 45 */
	unsigned char	track2Data[ISO1100_TRACK2_DATA_MAXLEN + 1];	/* BP 35 */
	Iso1100ProcCode	procCode;									/* BP 3 */
	unsigned char	acceptTermID[ISO1100_ACCEPT_TERM_ID_MAXLEN +1]; /* BP 41 */
	unsigned char	cardAcceptID[ISO1100_CARD_ACCEPT_ID_MAXLEN +1]; /* BP 42 */
	Iso1100MerchantIDType	cardAcceptIDType;					/* for BP 42 */
	unsigned char												/* BP 43 */
		cardAcceptLocationZip[ISO1100_CARD_ACCEPT_LOCATION_ZIP_MAXLEN + 1 ];
	unsigned char	systemTrace[ISO1100_SYSTEM_TRACE_MAXLEN + 1];	/* BP 11 */
	unsigned char	retrievalRef[ISO1100_RETRIEVAL_REF_MAXLEN + 1]; /* BP 37 */
	unsigned char	securityInfo[ISO1100_SECURITY_INFO_MAXLEN + 1];	/* BP 53 */
	unsigned char 	private1Data[ISO1100_PRIVATE1_DATA_MAXLEN + 1]; /* BP 62 */
	unsigned char	private2Data[ISO1100_PRIVATE2_DATA_MAXLEN + 1];	/* BP 63 */


	/* This is the bitmap that determines whether a particular data
	 * element has been set or not.  The size is defined by the 
	 * fact that there are a possible set of 64 data elements (as 
	 * defined by the ISO 8583 message).
	 */
	unsigned long			valueSetBitmap[8];

	/* This says whether or not the 'validate' method has been called
	 * and all the data necessary is valid for an Iso 8583 (1100) message
	 */
	unsigned int			objectValidated;

	Iso1100CardType			cardType;	/* What type of card is being
										 * validated (e.g. Visa, Amex).
										 */
} Iso1100Data;

/*
 * These are methods used by the Iso1100 object (in iso1100.c) but
 * defined in other modules.
 */

/*
 * Set methods 
 */
extern const Iso1100Status	Iso1100setAccountNo(Iso1100 *, 
										 const unsigned char*,
										 const Iso1100CharacterType);
extern const Iso1100Status	Iso1100setAmount(Iso1100 *, const float );
extern const Iso1100Status	Iso1100setAmountCurrency(Iso1100 *, 
											 const unsigned char*,
											 const Iso1100CharacterType);
extern const Iso1100Status	Iso1100setTransactionTime(Iso1100 *, 
													  const struct tm );
extern const Iso1100Status 	Iso1100setTransmissionTime(Iso1100 *, 
													   const struct tm );
extern const Iso1100Status	Iso1100setEffective(Iso1100 *, 
												const unsigned int year, 
												const unsigned int month);
extern const Iso1100Status	Iso1100setExpiration(Iso1100 *, 
												 const unsigned int year, 
												 const unsigned int month);
extern const Iso1100Status	Iso1100setMerchantType(Iso1100 *, 
											   const Iso1100MerchantType );
extern const Iso1100Status	Iso1100setCountry(Iso1100 *, 
											  const unsigned char*,
											  const Iso1100CharacterType);
extern const Iso1100Status	Iso1100setPosDataCode(Iso1100 *, 
										   const unsigned char*,
										   const Iso1100CharacterType);
extern const Iso1100Status	Iso1100setFunctionCode(Iso1100 *, 
												   const unsigned short );
extern const Iso1100Status	Iso1100setCardAccpBusCode(Iso1100 *, 
											   const Iso1100MerchantType );
extern const Iso1100Status	Iso1100setApprovalCodeLength(Iso1100 *, 
												  const unsigned short );
extern const Iso1100Status	Iso1100setAcquireInstID(Iso1100 *, 
												const unsigned char*,
											 const Iso1100CharacterType);
extern const Iso1100Status	Iso1100setForwardInstID(Iso1100 *, 
											 const unsigned char*,
											 const Iso1100CharacterType);
extern const Iso1100Status	Iso1100setTrack1Data(Iso1100 *, 
											 const unsigned char*,
										  const Iso1100CharacterType);
extern const Iso1100Status	Iso1100setTrack2Data(Iso1100 *, 
										  const unsigned char*,
										  const Iso1100CharacterType);
extern const Iso1100Status	Iso1100setProcCode(Iso1100 *, 
											   const Iso1100ProcCode );
extern const Iso1100Status	Iso1100setAcceptTermID(Iso1100 *, 
											const unsigned char*,
											const Iso1100CharacterType);
extern const Iso1100Status	Iso1100setCardAcceptID(Iso1100 *, 
											const Iso1100MerchantIDType,
											const unsigned char* id,
											const Iso1100CharacterType); 
extern Iso1100Status	Iso1100setCardAcceptLocation(Iso1100 *, 
											  const unsigned char* name,
											  const unsigned char* zip,
											  const unsigned char* region,
											  const unsigned char* country,
											  const Iso1100CharacterType);
extern Iso1100Status	Iso1100setSystemTrace(Iso1100 *, 
									   const unsigned char*,
									   const Iso1100CharacterType);
extern Iso1100Status	Iso1100setRetrievalRef(Iso1100 *, 
										const unsigned char*,
										const Iso1100CharacterType);
extern Iso1100Status	Iso1100setSecurityInfo(Iso1100 *, 
										const unsigned char*,
										const Iso1100CharacterType);
extern Iso1100Status	Iso1100setPrivate1Data(Iso1100 *, 
										const unsigned char*,
										const Iso1100CharacterType);
extern Iso1100Status	Iso1100setPrivate2Data(Iso1100 *, 
										const unsigned char*,
										const Iso1100CharacterType); 
/*
 * Get methods.
 */
extern const Iso1100Status Iso1100getRetrievalRef(Iso1100 *,
										   Iso1100CharacterType,
										   unsigned char *);

/*
 * Validate methods.
 */
const Iso1100Status Iso1100validateAccountNo(Iso1100*);
const Iso1100Status Iso1100validateAmount(Iso1100*);
const Iso1100Status Iso1100validateCurrency(Iso1100*);
const Iso1100Status Iso1100validateTransactionTime(Iso1100*);
const Iso1100Status Iso1100validateTransmissionTime(Iso1100*);
const Iso1100Status Iso1100validateEffectiveYear(Iso1100*);
const Iso1100Status Iso1100validateEffectiveMonth(Iso1100*);
const Iso1100Status Iso1100validateExpirationYear(Iso1100*);
const Iso1100Status Iso1100validateExpirationMonth(Iso1100*);
const Iso1100Status Iso1100validateMerchantType(Iso1100*);
const Iso1100Status Iso1100validateCountry(Iso1100*);
const Iso1100Status Iso1100validatePosDataCode(Iso1100*);
const Iso1100Status Iso1100validateFunctionCode(Iso1100*);
const Iso1100Status Iso1100validateCardAccpBusCode(Iso1100*);
const Iso1100Status Iso1100validateApprovalCodeLength(Iso1100*);
const Iso1100Status Iso1100validateAcquireInstId(Iso1100*);
const Iso1100Status Iso1100validateForwardInstId(Iso1100*);
const Iso1100Status Iso1100validateTrack1Data(Iso1100*);
const Iso1100Status Iso1100validateTrack2Data(Iso1100*);
const Iso1100Status Iso1100validateProcCode(Iso1100*);
const Iso1100Status Iso1100validateAcceptTermID(Iso1100*);
const Iso1100Status Iso1100validateCardAcceptID(Iso1100*);
const Iso1100Status Iso1100validateCardAcceptLocation(Iso1100*);
const Iso1100Status Iso1100validateSystemTrace(Iso1100*);
const Iso1100Status Iso1100validateRetrievalRef(Iso1100*);
const Iso1100Status Iso1100validateSecurityInfo(Iso1100*);
const Iso1100Status Iso1100validatePrivate1Data(Iso1100*);
const Iso1100Status Iso1100validatePrivate2Data(Iso1100*);

#endif /* !defined(__ISO1100INTERNAL_H) */
