#if !defined(__ISO1110_H)
#define __ISO1110_H
/**************************************************************************
 * File Name :		iso1110.h
 * 
 * Description : 	Declaration for class Iso1110
 *
 * Author : 		Ben Lurie
 *
 * Date Created : 	10/29/95
 *
 **************************************************************************
 */

/***************************************************************************
 *                      CLASS    Iso1110
 *
 * DESCRIPTION:
 *
 * The purpose of the Iso1110 class is to provide the ability to utilize
 * an Iso 8583 Authorization Response (1110) from a 'C' program.
 *
 * The Iso1100 class contains methods to get all of the fields as 
 * specified in the ISO 8583 1110 message (from the 
 * American Express Credit Authorization Guide).
 *
 * This object is only intended to be instantiated from an ISO 8583 
 * byte stream (as received from a back-end Amex server).  Therefore
 * it does not have any "set" methods, only "get" methods.
 *
 * NOTE:
 *
 *	All methods require passing a pointer to the ojbect to work on (this
 *	is necessary since C does not have real classes.
 *
 * This object is not thread-safe.
 *
 **************************************************************************
 */

#include <time.h>

#include "isomessage.h"

typedef IsoMessageStatus	Iso1110Status;
typedef IsoProcCode			Iso1110ProcCode;
typedef IsoCharacterType	Iso1110CharacterType;

/*
 * This type defines the information passed back as "additional response"
 * data (bit position 44).  Use the getAdditionalResponseData() method.
 */
typedef enum
{
	ADDRESS_AND_ZIP_CORRECT,
	ADDRESS_AND_ZIP_INCORRECT,
	ADDRESS_CORRECT,
	ZIP_CORRECT,
	INFORMATION_UNAVAIL,
	ADDRESS_VERIF_SUBSCR_VALID,
	SYSTEM_UNAVAIL
} Iso1110AddRespData;

/*
 * This type defines the action code for the authorization request (bit
 * position 39).  Use the getActionCode() method to access.
 */
typedef enum
{
	APPROVED = 0,
	APPROVE_WITH_ID = 1,
	APPROVE_VIP = 3,
	DENY = 100,
	EXPIRED_CARD = 101,
	DENY_NEW_CARD_ISSUED = 104,
	PLEASE_CALL_ISSUER = 107,
	INVALID_MERCHANT = 109,
	INVALID_AMOUNT = 110,
	INVALID_ACCOUNT = 111,
	SERVICE_NOT_PERMITTED = 115,
	INVALID_CID = 122,
	INVALID_EFFECTIVE_DATE = 125,
	FORMAT_ERROR = 181,
	PLEASE_WAIT = 182,
	INVALID_CURRENCY_CODE = 183,
	DENY_PICK_UP_CARD = 200,
	REVERSAL_ACCEPTED = 400
}	Iso1110ActionCode;

typedef struct Iso1110_t
{
	
	/************************************************************/
	/* Public methods.                                          */
	/************************************************************/

	/************************************************************/
	/* Get methods.                                           */
	/************************************************************/
	
	/* Bit Position 2 */
	const Iso1110Status (*getAccountNo)(struct Iso1110_t*,
										const Iso1110CharacterType,
										unsigned char *);

	/* Bit Position 3 */
	const Iso1110Status (*getProcCode)(struct Iso1110_t*,
									   Iso1110ProcCode*);

	/* Bit Position 4 */
	const Iso1110Status (*getAmount)(struct Iso1110_t*,
									 float*);
	
	/* Bit Position 7 */
	const Iso1110Status (*getTransmissionTime)(struct Iso1110_t*,
											   struct tm*);

	/* Bit Position 11 */
	const Iso1110Status (*getSystemTrace)(struct Iso1110_t*,
										  const Iso1110CharacterType,
										  unsigned char *);

	/* Bit Position 12 */
	const Iso1110Status (*getTransactionTime)(struct Iso1110_t*,
											  struct tm*);

	/* Bit Position 15 */
	const Iso1110Status (*getSettlement)(struct Iso1110_t*,
										 struct tm*);

	/* Bit Position 37 */
	const Iso1110Status (*getRetrievalRef)(struct Iso1110_t*, 
										   const Iso1110CharacterType,
										   unsigned char *);

	/* Bit Position 38 */
	const Iso1110Status (*getApprovalCode)(struct Iso1110_t*,
										   const Iso1110CharacterType,
										   unsigned char *);

	/* Bit Position 39 */
	const Iso1110Status (*getActionCode)(struct Iso1110_t*,
										 Iso1110ActionCode*);

	/* Bit Position 41 */
	const Iso1110Status (*getAcceptTermID)(struct Iso1110_t*,
										   const Iso1110CharacterType,	
										   unsigned char *);

	/* Bit Position 42 */
	const Iso1110Status (*getCardAcceptID)(struct Iso1110_t*,
										   const Iso1110CharacterType,
										   unsigned char *);
	
	/* Bit Position 44 */
	const Iso1110Status (*getAdditionalResponseData)(struct Iso1110_t*,
													 Iso1110AddRespData*);

	/* Bit Position 49 */
	const Iso1110Status (*getAmountCurrency)(struct Iso1110_t*,
											 const Iso1110CharacterType,
											 unsigned char *);

	/* Bit Position 62 */
	const Iso1110Status (*getPrivate1Data)(struct Iso1110_t*,
										   const Iso1110CharacterType,
										   unsigned char *);

	/* Bit Position 63 */
	const Iso1110Status (*getPrivate2Data)(struct Iso1110_t*,
										   const Iso1110CharacterType,
										   unsigned char *);

	/************************************************************/
	/* Other methods.                                           */
	/************************************************************/
	const Iso1110Status		(*status)(struct Iso1110_t *);
	const void*				(*getByteStream)(struct Iso1110_t *, size_t*);
	const unsigned char* 	(*getErrorString)(struct Iso1110_t *, 
											  const Iso1110Status);

	/************************************************************/
	/* Private methods and data.                                */
	/************************************************************/

	Iso1110Status			_status;

	/* This is the bitmap that determines whether a particular data
	 * element has been set or not.  The size is defined by the 
	 * fact that there are a possible set of 64 data elements (as 
	 * defined by the ISO 8583 message).
	 */
	unsigned long			valueBitmap[8];
	void*					byteStreamPtr;
	void*					dataPtr;

} Iso1110;

/********************************************************/
/* Maximum lengths for string data (not including a NULL
/* character at the end).
/********************************************************/


#define ISO1110_ACCOUNT_NO_BITPOSITION 2
#define ISO1110_ACCOUNT_NO_MAXLEN 19

#define ISO1110_PROCESSING_CODE_BITPOSITION 3
#define ISO1110_PROCESSING_CODE_MAXLEN 6

#define ISO1110_AMOUNT_BITPOSITION 4
#define ISO1110_AMOUNT_MAXLEN 12

#define ISO1110_DATE_TIME_TRANSMISSION_BITPOSITION 7
#define ISO1110_DATE_TIME_TRANSMISSION_MAXLEN 10

#define ISO1110_SYSTEM_TRACE_BITPOSITION 11
#define ISO1110_SYSTEM_TRACE_MAXLEN 6

#define ISO1110_DATE_TIME_TRANSACTION_BITPOSITION 12
#define ISO1110_DATE_TIME_TRANSACTION_MAXLEN 12

#define ISO1110_SETTLEMENT_BITPOSITION 15
#define ISO1110_SETTLEMENT_MAXLEN 6

#define ISO1110_RETRIEVAL_REF_BITPOSITION 37
#define ISO1110_RETRIEVAL_REF_MAXLEN 12

#define ISO1110_APPROVAL_CODE_BITPOSITION 38
#define ISO1110_APPROVAL_CODE_MAXLEN 6

#define ISO1110_ACTION_CODE_BITPOSITION 39
#define ISO1110_ACTION_CODE_MAXLEN 3

#define ISO1110_ACCEPT_TERM_ID_BITPOSITION 41
#define ISO1110_ACCEPT_TERM_ID_MAXLEN 8

#define ISO1110_CARD_ACCEPT_ID_BITPOSITION 42
#define ISO1110_CARD_ACCEPT_ID_MAXLEN 15

#define ISO1110_ADDITIONAL_RESPONSE_DATA_BITPOSITION 44
#define ISO1110_ADDITIONAL_RESPONSE_DATA_MAXLEN 25

#define ISO1110_CURRENCY_CODE_BITPOSITION 49
#define ISO1110_CURRENCY_CODE_MAXLEN 3

#define ISO1110_PRIVATE1_DATA_BITPOSITION 62
#define ISO1110_PRIVATE1_DATA_MAXLEN 60

#define ISO1110_PRIVATE2_DATA_BITPOSITION 63
#define ISO1110_PRIVATE2_DATA_MAXLEN 100

/********************************************************/
/* Maximum length of any bitposition.
/********************************************************/
#define ISO1110_BITPOSITION_MAXLEN		(100)

/********************************************************/
/* Instantiate an instance of an ISO 1110 message.
/********************************************************/

extern Iso1110*	initIso1110(void);
extern Iso1110* initIso1110FromByteStream(const void *, const size_t numBytes);
extern void 	deleteIso1110(Iso1110*);

#endif /* !defined(__ISO1110_H) */

