#if !defined(__ISO1110INTERNAL_H)
#define __ISO1110INTERNAL_H

/**************************************************************************
 * File Name :		iso1110internal.h
 * 
 * Description : 	Internal class definitions for the Iso1110 object
 *
 * Author : 		Ben Lurie
 *
 * Date Created : 	10/29/95
 *
 **************************************************************************
 */

#include "iso1110.h"

/********************************************************/
/* Preprocessor defines 								*/
/********************************************************/

/**************************************************************************
 *
 * NAME:                valueSet
 *
 * ARGUMENTS:           thisPtr
 *							Pointer to Iso1110 message to work on.
 *
 *						bitPosition
 *							Bitposition of data that has been set.
 *
 * RETURN VALUE:		None.
 *
 * SCOPE:               private
 *
 * DESCRIPTION:
 *
 * This macro records the fact that some piece of
 * data, whose bitposition is 'bitPosition', has been set in the
 * internal data structure. 
 *
 * The current implementation of this is to set a bit in the 
 * 'valueBitmap' field.
 *
 **************************************************************************
 */
#define valueSet(thisPtr, bitPosition) \
	IsoSetBit(thisPtr->valueBitmap, bitPosition)

/**************************************************************************
 *
 * NAME:                valueUnset
 *
 * ARGUMENTS:           thisPtr
 *							Pointer to Iso1110 message to work on.
 *
 *						bitPosition
 *							Bitposition of data that has been unset.
 *
 * RETURN VALUE:		None.
 *
 * SCOPE:               private
 *
 * DESCRIPTION:
 *
 * This macro records the fact that some piece of
 * data, whose bitposition is 'bitPosition', has been unset from the
 * internal data structure. 
 *
 * The current implementation of this is to unset a bit in the 
 * 'valueBitmap' field.
 *
 **************************************************************************
 */
#define valueUnset(thisPtr, bitPosition) \
	IsoClearBit(thisPtr->valueBitmap, bitPosition)

/**************************************************************************
 *
 * NAME:                valueIsSet
 *
 * ARGUMENTS:           thisPtr
 *							Pointer to Iso1110 message to work on.
 *
 *						bitPosition
 *							Bitposition of data to be tested.
 *
 * RETURN VALUE:		TRUE if data at 'bitPosition' has been set,
 *						FALSE if data has not been set.
 *
 * SCOPE:               private
 *
 * DESCRIPTION:
 *
 * This macro tests to see if data at a particular bitposition has
 * been set in the internal data structure.
 *
 * The current implementation of this is to test a bit in the 
 * 'valueBitmap' field.
 *
 **************************************************************************
 */
#define valueIsSet(thisPtr, bitPosition) \
	IsoTestBit(thisPtr->valueBitmap, bitPosition)

/**************************************************************************
 *
 * NAME:                isObjectValid
 *
 * ARGUMENTS:           thisPtr
 *							Pointer to Iso1110 message to work on.
 *
 * RETURN VALUE:		TRUE if object can be used, 
 *						FALSE if object is not valid to use.
 *
 * SCOPE:               private
 *
 * DESCRIPTION:
 *
 * This macro tests to see if the object is valid to use.  If the
 * status is 'INVALID' or if there has been a problem with memory,
 * then the object is not valid to use.
 *
 * This is used by the methods of the Iso1110 object to see if 
 * they are being passed a valid object.
 *
 **************************************************************************
 */
#define isObjectValid(thisPtr) \
	((thisPtr->_status != INVALID) && (thisPtr->_status != ENOMEM))

/*
 * This is the internal data of an Iso1110 message.  
 *
 * It contains one data element for each bit position that can 
 * be set in an Iso1110 message.  The data element types are 
 * typically influenced by how the user is going to get the information
 * from the 'get' method (and this is influenced by the fact that this
 * information is going to be retrieved programatically from a 'C' program).
 *
 * NOTE:
 * 		The main purpose of having the data elements reside in the
 *		Iso1110 message is for:
 *			
 *			*	efficient retrieval.  We don't have to get the information
 *				from the Iso byte stream object and reformat it for
 *				the user.
 *
 * 		This is basically a time versus space tradeoff.
 */
typedef struct Iso1110Data_t
{
	unsigned char	accountNo[ISO1110_ACCOUNT_NO_MAXLEN + 1];	/* BP 2 */
	Iso1110ProcCode	procCode;									/* BP 3 */
	float			amount;										/* BP 4 */
	struct tm		transmissionTime;							/* BP 7 */
	unsigned char	systemTrace[ISO1110_SYSTEM_TRACE_MAXLEN + 1];	/* BP 11 */
	struct tm		transactionTime;							/* BP 12 */
	struct tm		settlement;									/* BP 15 */
	unsigned char	retrievalRef[ISO1110_RETRIEVAL_REF_MAXLEN + 1]; /* BP 37 */
	unsigned char	approvalCode[ISO1110_APPROVAL_CODE_MAXLEN + 1]; /* 38 */
	Iso1110ActionCode actionCode;								/* BP 39 */
	unsigned char	acceptTermID[ISO1110_ACCEPT_TERM_ID_MAXLEN +1]; /* BP 41 */
	unsigned char	cardAcceptID[ISO1110_CARD_ACCEPT_ID_MAXLEN +1]; /* BP 42 */
	Iso1110AddRespData additionalResponseData;						/* BP 44 */
	unsigned char	currencyCode[ISO1110_CURRENCY_CODE_MAXLEN + 1]; /* BP 49 */
	unsigned char 	private1Data[ISO1110_PRIVATE1_DATA_MAXLEN + 1]; /* BP 62 */
	unsigned char	private2Data[ISO1110_PRIVATE2_DATA_MAXLEN + 1];	/* BP 63 */
} Iso1110Data;

/*
 * These are methods used by the Iso1110 object (in iso1110.c) but
 * defined in other modules.
 */

/* 
 * Get methods
 */

/* Bit Position 2 */
extern const Iso1110Status Iso1110getAccountNo(Iso1110*,
										const Iso1110CharacterType,
										unsigned char *);

/* Bit Position 3 */
extern const Iso1110Status Iso1110getProcCode(Iso1110*,
									   Iso1110ProcCode*);

/* Bit Position 4 */
extern const Iso1110Status Iso1110getAmount(Iso1110*,
									 float*);

/* Bit Position 7 */
extern const Iso1110Status Iso1110getTransmissionTime(Iso1110*,
											   struct tm*);

/* Bit Position 11 */
extern const Iso1110Status Iso1110getSystemTrace(Iso1110*,
										  const Iso1110CharacterType,
										  unsigned char *);

/* Bit Position 12 */
extern const Iso1110Status Iso1110getTransactionTime(Iso1110*,
											  struct tm*);

/* Bit Position 15 */
extern const Iso1110Status Iso1110getSettlement(Iso1110*,
										 struct tm*);

/* Bit Position 37 */
extern const Iso1110Status Iso1110getRetrievalRef(Iso1110*, 
										   const Iso1110CharacterType,
										   unsigned char *);

/* Bit Position 38 */
extern const Iso1110Status Iso1110getApprovalCode(Iso1110*,
										   const Iso1110CharacterType,
										   unsigned char *);

/* Bit Position 39 */
extern const Iso1110Status Iso1110getActionCode(Iso1110*,
										 Iso1110ActionCode*);

/* Bit Position 41 */
extern const Iso1110Status Iso1110getAcceptTermID(Iso1110*,
										   const Iso1110CharacterType,	
										   unsigned char *);

/* Bit Position 42 */
extern const Iso1110Status Iso1110getCardAcceptID(Iso1110*,
										   const Iso1110CharacterType,
										   unsigned char *);


/* Bit Position 44 */
extern const Iso1110Status Iso1110getAdditionalResponseData(Iso1110*,
													 Iso1110AddRespData*);

/* Bit Position 49 */
extern const Iso1110Status Iso1110getAmountCurrency(Iso1110*,
													const Iso1110CharacterType,
													unsigned char *);

/* Bit Position 62 */
extern const Iso1110Status Iso1110getPrivate1Data(Iso1110*,
										   const Iso1110CharacterType,
										   unsigned char *);

/* Bit Position 63 */
extern const Iso1110Status Iso1110getPrivate2Data(Iso1110*,
										   const Iso1110CharacterType,
										   unsigned char *);
#endif /* !defined(__ISO1110INTERNAL_H) */
