/************************************************************
 *
 *     order.c
 *
 *  API for vendor-specific code in transaction daemon
 *     
 * RMR - 06/05/96 -  changed read_len calculation - ACTION_CODE
 *   and APPROVAL_CODE are new fields inthe ORDER HEADER record
 ************************************************************/
#include "ccamexserverutil.h"
#include "iso1100.h"
#include "iso1110.h"
#include "ccamexauthserver.h"

#include "transd.h"
#include "order.h"

struct AMEXOrd {
  char                  acct_no[16];
  Iso1100ProcCode       proc_code;
  float                 amount;
  unsigned char         trace[ORDER_NUM_LEN];  /* order number */
  unsigned int          exp_month;
  unsigned int          exp_year;
  Iso1100MerchantType   merchant_type;
  unsigned char         country[4];
  unsigned char         pos_data[13];         /* AMEX provides */
  Iso1100MerchantType   card_acceptor_bus_code;
  unsigned int          approval_code_len;
  unsigned char         card_accept_id[11];   /* AMEX provides */
  struct CardAcceptLoc  acceptor_loc;       /* AMEX provides */
  unsigned char         currency_code[4];
  unsigned char         security_info[8];      /* AMEX provides */
  Iso1110ActionCode     action_code;
  char                  approval_code[APPROVAL_CODE_LEN+1];
};

/*---------------------------------------------------------------------------*/
/* function declarations                                                     */
/*---------------------------------------------------------------------------*/
extern int  amex_auth(struct AMEXOrd *amex_order);
int         open_lock_file(char*, int);
int         unlock_close_file(int, char*);
void        strip_blanks(char*);
void        get_field(char*, char*, int, int);
int         print_numeric_field(int, char*, int);
int         print_char_field(int, char*, int);

/************************************************************
 *
 *  OrderInit()
 *
 *	This function is called from the main thread before the
 *	main thread enters its dispatch loop. The vendor may
 *	add initialization code here, e.g., mutex initialization
 *	for multi-threaded database access.
 *
 ************************************************************/

static transd_errvec errmsg;

int OrderInit(void)
{
  logger(LOG_DEBUG, "%s", "Initializing vendor order module");

  errmsg[0] = "Thanks for buying from BOC Gases!";
  errmsg[TRANSD_ORDER_FILE_ERROR] = "Error updating order with American Express authorization info.";
  errmsg[TRANSD_NOSUCH_ORDER] = "Could not find part of or whole order.";
  errmsg[TRANSD_USER_WRONG] = "Your username does not match the username on the specified order.";
  errmsg[TRANSD_ALREADY_SIGNED] = "Order is already signed.";
  errmsg[TRANSD_AMEXAUTH_ERROR] = "Error trying to get American Express authorization for charge.";

  return 0;
}


/************************************************************
 *
 *  OrderText()
 *
 *	This function returns the ascii order text for a given
 *	order number. This sample code assumes that the order
 *	text is stored as a file named "orderno" in the current
 *	working directory -- this is a very simple database!
 *	The vendor should replace this sample implementation
 *	with access to the vendor order database.
 *
 *	Note the following:
 *
 *	o the text or error string returned must be malloc'ed,
 *		since the main will attempt to free it, after
 *		sending it back to the requestor.
 *
 *	o it is the responsibility of the vendor to check that
 *		the "user" is the creator of the order.
 *
 *	o the environment is multi-threaded. This will be
 *		completely transparent so long as the usual
 *		programming rules are followed: e.g., no static
 *		variables, don't call non thread-safe routines,
 *		etc. If these rules cannot be followed, then
 *		mutexes and condition variables must be used
 *		as required.
 *
 ************************************************************/

long OrderText(char *orderno, char *user, char **pmsg)
{
  int    file_desc, open_flag, found_order, total_items, read_len, num_bytes;
  int    field_len, field_start, num_items, index, num_diff_items, num_accts, user_len;
  char   c_total_items[32];
  char   *msg;
  long   status;
  char   buff[COMMENTS_LEN+1];
  char   file_ordrnum[ORDER_NUM_LEN];
  char   record_type[RECORD_TYPE_LEN];
  char   order_total[ORDER_TOTAL_LEN];
  char   order_status[2];
  char   comments[COMMENTS_LEN];
  char   file_username[USERID_LEN];
  char   req_name[NAME_LEN * 2];
  char   phone[PHONE_LEN];
  char   email_addr[ ADDRESS_LEN];
  char   ship_name[NAME_LEN * 2];
  char   ship_addr[ADDRESS_LEN * 2];
  char   city[ADDRESS_LEN];
  char   state[STATE_LEN];
  char   zip[ZIP_LEN];
  char   account[ACCOUNT_LEN];
  char   percent[PERCENT_LEN];
  p_acct first_acct, new_acct, previous_acct;
  p_item first_item, new_item, previous_item;

  /* initialize values for parsing orders file */
  status = 0;
  open_flag = 0;
  found_order = 0;
  total_items = 0;
  num_accts = 0;
  num_diff_items = 0;
  first_acct = NULL;
  first_item = NULL;

  /* open and lock the orders file */
  if ((file_desc = open_lock_file(DBNAME, O_RDWR)) == -1)
  {
    status = TRANSD_ORDER_FILE_ERROR;
    goto Exit;
  }
  else
    open_flag = 1;

  /* rewind file before reading from it */
  lseek(file_desc, 0L, 0);

  /* read first order number and record type from orders file */
  read_len = ORDER_NUM_LEN + RECORD_TYPE_LEN - 2;
  num_bytes = read(file_desc, buff, read_len);
  buff[num_bytes] = '\0'; 

  /* for all lines in orders file until find whole order */
  while (num_bytes != 0 && found_order == 0)
  {
    if (num_bytes == -1)
    {
      logger(LOG_ERR, "%s %s %s", "Error reading the", DBNAME, "file.");
      status = TRANSD_ORDER_FILE_ERROR;
      goto Exit;
    }

    /* get order number from file buffer starting in col 0 of buff */
    field_len = ORDER_NUM_LEN - 1;
    field_start = 0;
    get_field(buff, file_ordrnum, field_start, field_len);

    /* get record type from buff */
    field_len = RECORD_TYPE_LEN - 1;
    field_start = ORDER_NUM_LEN - 1;
    get_field(buff, record_type, field_start, field_len);

    /* if record type = HEADER_RECTYPE, process order header line */
    if (strcmp(record_type, HEADER_RECTYPE) == 0)
    {
      /* read in rest of order header line */
      read_len = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 4
                 + PHONE_LEN + ADDRESS_LEN * 4 + STATE_LEN + ZIP_LEN
                 + ACCOUNT_LEN * 3 + PERCENT_LEN * 3 + CC_NUM_LEN
                 + CC_EXP_LEN + ACTION_CODE_LEN + APPROVAL_CODE_LEN - 22;
      num_bytes = read(file_desc, buff, read_len);
      buff[num_bytes] = '\0';

      /* if line is for the current order number, get order info */
      if (strcmp(file_ordrnum, orderno) == 0)
      {
        /* get username from buff */
        field_len = USERID_LEN - 1;
        field_start = DATE_TIME_LEN - 1;
        get_field(buff, file_username, field_start, field_len);

        /* verify "user" is the owner of this order */
        user_len = strlen(file_username);
        if (strncmp(file_username, user, user_len) != 0)
        {
          status = TRANSD_USER_WRONG;
          goto Exit;
        }

        /* get requisitioner name from buff */
        field_len = (NAME_LEN - 1) * 2;
        field_start = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN - 3;
        get_field(buff, req_name, field_start, field_len);

        /* get email address from buff */
        field_len = ADDRESS_LEN - 1;
        field_start = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 2 - 5;
        get_field(buff, email_addr, field_start, field_len);

        /* get phone number from buff */
        field_len = PHONE_LEN - 1;
        field_start = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 2 + ADDRESS_LEN - 6;
        get_field(buff, phone, field_start, field_len);

        /* get ship to name from buff */
        field_len = (NAME_LEN - 1) * 2;
        field_start = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 2 + ADDRESS_LEN + PHONE_LEN - 7;
        get_field(buff, ship_name, field_start, field_len);

        /* get ship to address from buff */
        field_len = (ADDRESS_LEN - 1) * 2;
        field_start = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 4 + ADDRESS_LEN + PHONE_LEN - 9;
        get_field(buff, ship_addr, field_start, field_len);

        /* get ship to city from buff */
        field_len = ADDRESS_LEN - 1;
        field_start = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 4 + ADDRESS_LEN * 3 + PHONE_LEN - 11;
        get_field(buff, city, field_start, field_len);

        /* get ship to state from buff */
        field_len = STATE_LEN - 1;
        field_start = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 4 + ADDRESS_LEN * 4 + PHONE_LEN - 12;
        get_field(buff, state, field_start, field_len);

        /* get ship to zip code from buff */
        field_len = ZIP_LEN - 1;
        field_start = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 4 + ADDRESS_LEN * 4 + PHONE_LEN + STATE_LEN - 13;
        get_field(buff, zip, field_start, field_len);

        for (index = 0; index < MAX_ACCOUNTS; index++)
        {
          /* get account number */
          field_len = ACCOUNT_LEN - 1;
          field_start = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN
            + NAME_LEN * 4 + ADDRESS_LEN * 4 + PHONE_LEN + STATE_LEN
            + ZIP_LEN + index * (ACCOUNT_LEN + PERCENT_LEN - 2) - 14;
          get_field(buff, account, field_start, field_len);

          /* get percentage to charge to account */
          field_len = PERCENT_LEN - 1;
          field_start = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN
            + NAME_LEN * 4 + ADDRESS_LEN * 4 + PHONE_LEN + STATE_LEN
            + ZIP_LEN + ACCOUNT_LEN + index * (ACCOUNT_LEN + PERCENT_LEN - 2) - 15;
          get_field(buff, percent, field_start, field_len);

          /* increment number of accounts and add to list if account not None */
          if (strcmp(account, "None") != 0)
          {
            num_accts++;

            /* allocate structure for new account in list */
            if ((new_acct = (p_acct) malloc(sizeof(Acct))) == NULL)
            {
              logger(LOG_ERR, "%s", "Memory allocation error for account.");
              status = TRANSD_ORDER_FILE_ERROR;
              goto Exit;
            }

            /* set new account values */
            strcpy(new_acct->number, account);
            strcpy(new_acct->percent, percent);
            new_acct->next_acct = NULL;

            /* set pointers to put new account at end of list */
            if (first_acct == NULL)
            {
              first_acct = new_acct;
              previous_acct = new_acct;
            }
            else
            {
              previous_acct->next_acct = new_acct;
              previous_acct = new_acct;
            }
          }
        }
      }
    }
    /* if record type = COMMENT_RECTYPE, process order comment line */
    else if (strcmp(record_type, COMMENT_RECTYPE) == 0)
    {
      /* read in rest of order comments line */
      read_len = COMMENTS_LEN;
      num_bytes = read(file_desc, buff, read_len);
      buff[num_bytes] = '\0';

      /* if line is for the current order number, get order info */
      if (strcmp(file_ordrnum, orderno) == 0)
      {
        /* get item code from buff */
        field_len = COMMENTS_LEN - 1;
        field_start = 0;
        get_field(buff, comments, field_start, field_len);
      }
    }
    /* if record type = DETAIL_RECTYPE, process order detail line */
    else if (strcmp(record_type, DETAIL_RECTYPE) == 0)
    {
      /* read in rest of order detail line */
      read_len = ITEM_CODE_LEN + QUANTITY_LEN * 2 + PRICE_LEN + UNIT_LEN - 4;
      num_bytes = read(file_desc, buff, read_len);
      buff[num_bytes] = '\0';

      /* if line is for the current order number, get order info */
      if (strcmp(file_ordrnum, orderno) == 0)
      {
        /* allocate structure for new account in list */
        if ((new_item = (p_item) malloc(sizeof(Item))) == NULL)
        {
          logger(LOG_ERR, "%s", "Memory allocation error for item.");
          status = TRANSD_ORDER_FILE_ERROR;
          goto Exit;
        }

        /* get item code from buff */
        field_len = ITEM_CODE_LEN - 1;
        field_start = 0;
        get_field(buff, new_item->code, field_start, field_len);

        /* get # orders from buff */
        field_len = QUANTITY_LEN - 1;
        field_start = ITEM_CODE_LEN - 1;
        get_field(buff, new_item->orders, field_start, field_len);

        /* get # returns from buff */
        field_start = ITEM_CODE_LEN + QUANTITY_LEN - 2;
        get_field(buff, new_item->returns, field_start, field_len);

        /* get item price from buff */
        field_len = PRICE_LEN - 1;
        field_start = ITEM_CODE_LEN + QUANTITY_LEN * 2 - 3;
        get_field(buff, new_item->price, field_start, field_len);

        /* calculate values and totals */
        num_diff_items++;
        num_items = atoi(new_item->orders);
        total_items += num_items;

        /* set pointers to put new item at end of list */
        new_item->next_item = NULL;
        if (first_item == NULL)
        {
          first_item = new_item;
          previous_item = new_item;
        }
        else
        {
          previous_item->next_item = new_item;
          previous_item = new_item;
        }
      }
    }
    /* if record type = TRAILER_RECTYPE, process order trailer line */
    else if (strcmp(record_type, TRAILER_RECTYPE) == 0)
    {
      /* read in rest of order trailer line */
      read_len = ORDER_TOTAL_LEN + 1;
      num_bytes = read(file_desc, buff, read_len);
      buff[num_bytes] = '\0';

      /* if line is for the current order number, get order info */
      if (strcmp(file_ordrnum, orderno) == 0)
      {
        /* get order total from buff */
        field_len = ORDER_TOTAL_LEN - 1;
        field_start = 0;
        get_field(buff, order_total, field_start, field_len);

        /* get order status from buff */
        field_len = 1;
        field_start = ORDER_TOTAL_LEN - 1;
        get_field(buff, order_status, field_start, field_len);

        /* set done reading specified order flag to true */
        found_order = 1;
      }
    }
    /* if record type not valid, error */
    else
    {
      logger(LOG_ERR, "%s", "Invalid record type found in orders file.");
      status = TRANSD_ORDER_FILE_ERROR;
      goto Exit;
    }

    /* get next order number and record type from order file */
    read_len = ORDER_NUM_LEN + RECORD_TYPE_LEN - 2;
    num_bytes = read(file_desc, buff, read_len);
    buff[num_bytes] = '\0'; 
  } /* end of code to read in lines of orders file */

  /* check if complete order found */
  if (found_order == 0)
  {
    status = TRANSD_NOSUCH_ORDER;
    goto Exit;
  }

  /* convert total items count to character string */
  sprintf(c_total_items, "%d", total_items);

  /* determine the number of bytes to malloc for this order */
  num_bytes = strlen("Order #\n") + strlen(orderno);
  num_bytes += strlen("Requisitioner: , \n") + strlen(req_name) + strlen(user);
  num_bytes += strlen("Requisitioner Email Address: \n") + strlen(email_addr);
  num_bytes += strlen("Requisitioner Phone Number: \n") + strlen(phone);
  num_bytes += strlen("Ship-To Name: \n") + strlen(ship_name);
  num_bytes += strlen("Ship-To Address: \n") + strlen(ship_addr);
  num_bytes += strlen("City: , State:  Zip: \n") + strlen(city) + strlen(state) + strlen(zip) + strlen("\n");
  num_bytes += strlen("Accounts Charged: \n");
  num_bytes += (strlen("Account #: ,  percent\n") + ACCOUNT_LEN + PERCENT_LEN) * num_accts + strlen("\n");
  num_bytes += strlen("Order Comments: \n") + strlen(comments) + strlen("\n");
  num_bytes += strlen("Items Purchased: \n");
  num_bytes += (strlen("Iem Code: , #Items: , #Returns: , Price: \n") + ITEM_CODE_LEN + QUANTITY_LEN * 2 + PRICE_LEN) * num_diff_items + strlen("\n");
  num_bytes += strlen("You just purchased  line item(s) totaling $.\n") + strlen(c_total_items) + strlen(order_total);

  /* malloc space */
  msg = (char *) malloc(num_bytes);

  /* add order text onto msg */
  strcpy(msg, "\0");
  strcat(msg, "Order #");
  strcat(msg, orderno);
  strcat(msg, "\nRequisitioner: ");
  strcat(msg, req_name);
  strcat(msg, ", ");
  strcat(msg, user);
  strcat(msg, "\nRequisitioner Email Address: ");
  strcat(msg, email_addr);
  strcat(msg, "\nRequisitioner Phone Number: ");
  strcat(msg, phone);
  strcat(msg, "\nShip-To Name: ");
  strcat(msg, ship_name);
  strcat(msg, "\nShip-To Address: ");
  strcat(msg, ship_addr);
  strcat(msg, "\nCity: ");
  strcat(msg, city);
  strcat(msg, ", State: ");
  strcat(msg, state);
  strcat(msg, " Zip: ");
  strcat(msg, zip);
  strcat(msg, "\n\nAccounts Charged: ");
  new_acct = first_acct;
  while (new_acct != NULL)
  {
    strcat(msg, "\nAccount #: ");
    strcat(msg, new_acct->number);
    strcat(msg, ", ");
    strcat(msg, new_acct->percent);
    strcat(msg, " percent");
    new_acct = new_acct->next_acct;
  }
  strcat(msg, "\n\nOrder Comments: ");
  strcat(msg, comments);
  strcat(msg, "\n\nItems Purchased: ");
  new_item = first_item;
  while (new_item != NULL)
  {
    strcat(msg, "\nItem Code: ");
    strcat(msg, new_item->code);
    strcat(msg, ", #Items: ");
    strcat(msg, new_item->orders);
    strcat(msg, ", #Returns: ");
    strcat(msg, new_item->returns);
    strcat(msg, ", Price: ");
    strcat(msg, new_item->price);
    new_item = new_item->next_item;
  }
  strcat(msg, "\n\nYou just purchased ");
  strcat(msg, c_total_items);
  strcat(msg, " line item(s) totaling $");
  strcat(msg, order_total);
  strcat(msg, ".\n");

  /* free space malloc'ed for account and item order info */
  while (first_item != NULL)
  {
    new_item = first_item->next_item;
    free(first_item);
    first_item = new_item;
  }

  while (first_acct != NULL)
  {
    new_acct = first_acct->next_acct;
    free(first_acct);
    first_acct = new_acct;
  }

  /* close the order file and set return pointer */
  unlock_close_file(file_desc, DBNAME);
  *pmsg = msg;
  return status;

  /* Exit gracefully */
  Exit:
  {
    while (first_item != NULL)
    {
      new_item = first_item->next_item;
      free(first_item);
      first_item = new_item;
    }

    while (first_acct != NULL)
    {
      new_acct = first_acct->next_acct;
      free(first_acct);
      first_acct = new_acct;
    }

    if (open_flag == 1)
      unlock_close_file(file_desc, DBNAME);

    msg = (char*) malloc(strlen(errmsg[status])+1);
    strcpy(msg, errmsg[status]);
    *pmsg = msg;

    return status;
  }
} /* OrderText */


/************************************************************
 *
 *  OrderAccept()
 *
 *	This function changes the status of an order in the
 *	vendor order database to "accepted" or "signed". This
 *	sample code does nothing, other than verifying the
 *	existence of the order.
 *
 *	The programming comments in the note preceding OrderText()
 *	apply here as well.
 *
 ************************************************************/

long OrderAccept(char *orderno, char *user, char **pmsg)
{
  int    file_desc, read_len, num_bytes, open_flag, found_order;
  int    field_len, field_start, user_len, rc;
  double order_total;
  long   status;
  char   *msg, **index_ptr, **sprofile;
  char   ccnum[CC_NUM_LEN];
  char   ccexp[CC_EXP_LEN];
  char   ccexp_year[3];
  char   ccexp_month[3];
  char   buff[COMMENTS_LEN+1];
  char   file_ordrnum[ORDER_NUM_LEN];
  char   record_type[RECORD_TYPE_LEN];
  char   c_order_total[ORDER_TOTAL_LEN];
  char   order_status[2];
  char   file_username[USERID_LEN];
  struct AMEXOrd AMEXdata;

  /* initialize values for parsing orders file */
  status = 0;
  open_flag = 0;
  found_order = 0;

  /* get user's credit card # and expiration date from secure profile database */
  msg = GetSecureProfile(user, &sprofile);

  if (!msg)
  {
    index_ptr = sprofile;
    strcpy(ccnum, "\0");
    strcpy(ccexp, "\0");

    /* get to beginning of procard number keyword */
    while (strcmp(*index_ptr, "procard_number") != 0)
      *index_ptr += 1;

    /* get to beginning of procard number (after keyword) */
    *index_ptr += strlen(*index_ptr);
    *index_ptr += 1;

    /* set procard number */
    strcpy(ccnum, *index_ptr);

    /* go to end of procard number (beginning of procard exp date keyword) */
    *index_ptr += strlen(*index_ptr);
    *index_ptr += 1;

    /* get to beginning of procard expiration date (after keyword) */
    *index_ptr += strlen(*index_ptr);
    *index_ptr += 1;

    /* set procard expiration date */
    strcpy(ccexp, *index_ptr);

    free(sprofile);
  }
  else
  {
    logger(LOG_ERR, "%s", msg);
    errmsg[TRANSD_GETSPROF_ERROR] = msg;
    free(sprofile);
    status = TRANSD_GETSPROF_ERROR;
    goto Exit;
  }

  /* open and lock the orders file */
  if ((file_desc = open_lock_file(DBNAME, O_RDWR)) == -1)
  {
    status = TRANSD_ORDER_FILE_ERROR;
    goto Exit;
  }
  else
    open_flag = 1;

  /* rewind file before reading from it */
  lseek(file_desc, 0L, 0);

  /* read first order number and record type from orders file */
  read_len = ORDER_NUM_LEN + RECORD_TYPE_LEN - 2;
  num_bytes = read(file_desc, buff, read_len);
  buff[num_bytes] = '\0'; 

  /* for all lines in orders file */
  while (num_bytes != 0)
  {
    if (num_bytes == -1)
    {
      logger(LOG_ERR, "%s %s %s", "Error reading the", DBNAME, "file.");
      status = TRANSD_ORDER_FILE_ERROR;
      goto Exit;
    }

    /* get order number from file buffer starting in col 0 of buff */
    field_len = ORDER_NUM_LEN - 1;
    field_start = 0;
    get_field(buff, file_ordrnum, field_start, field_len);

    /* get record type from buff */
    field_len = RECORD_TYPE_LEN - 1;
    field_start = ORDER_NUM_LEN - 1;
    get_field(buff, record_type, field_start, field_len);

    /* if record type = HEADER_RECTYPE, process order header line */
    if (strcmp(record_type, HEADER_RECTYPE) == 0)
    {
      /* if line is for current order number, replace credit card number and exp date */
      if (strcmp(file_ordrnum, orderno) == 0)
      {
        /* read in part of order header line */
        read_len = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 4
                   + PHONE_LEN + ADDRESS_LEN * 4 + STATE_LEN + ZIP_LEN
                   + ACCOUNT_LEN * 3 + PERCENT_LEN * 3 - 20;
        num_bytes = read(file_desc, buff, read_len);
        buff[num_bytes] = '\0';

        /* get username from buff */
        field_len = USERID_LEN - 1;
        field_start = DATE_TIME_LEN - 1;
        get_field(buff, file_username, field_start, field_len);

        /* verify "user" is the owner of this order */
        user_len = strlen(file_username);
        if (strncmp(file_username, user, user_len) != 0)
        {
          status = TRANSD_USER_WRONG;
          goto Exit;
        }

        /* write credit card number */
        if (print_numeric_field(file_desc, ccnum, CC_NUM_LEN) == 0)
        {
          logger(LOG_ERR, "%s", "Error writing credit card number to file.");
          status = TRANSD_ORDER_FILE_ERROR;
          goto Exit;
        }

        /* write credit card expiration date */
        if (print_numeric_field(file_desc, ccexp, CC_EXP_LEN) == 0)
        {
          logger(LOG_ERR, "%s", "Error writing credit card expiration date to file.");
          status = TRANSD_ORDER_FILE_ERROR;
          goto Exit;
        }

        /* read in rest of order header line */
        read_len = ACTION_CODE_LEN + APPROVAL_CODE_LEN;
        num_bytes = read(file_desc, buff, read_len);
        buff[num_bytes] = '\0';
      }
      /* else read in rest of order header line */
      else
      {
        read_len = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 4
                   + PHONE_LEN + ADDRESS_LEN * 4 + STATE_LEN + ZIP_LEN
                   + ACCOUNT_LEN * 3 + PERCENT_LEN * 3 + CC_NUM_LEN
                   + CC_EXP_LEN + ACTION_CODE_LEN + APPROVAL_CODE_LEN - 22;
        num_bytes = read(file_desc, buff, read_len);
        buff[num_bytes] = '\0';
      }
    }
    /* if record type = COMMENT_RECTYPE, process order comment line */
    else if (strcmp(record_type, COMMENT_RECTYPE) == 0)
    {
      /* read in rest of order comments line */
      read_len = COMMENTS_LEN;
      num_bytes = read(file_desc, buff, read_len);
      buff[num_bytes] = '\0';
    }
    /* if record type = DETAIL_RECTYPE, process order detail line */
    else if (strcmp(record_type, DETAIL_RECTYPE) == 0)
    {
      /* read in rest of order detail line */
      read_len = ITEM_CODE_LEN + QUANTITY_LEN * 2 + PRICE_LEN + UNIT_LEN - 4;
      num_bytes = read(file_desc, buff, read_len);
      buff[num_bytes] = '\0';
    }
    /* if record type = TRAILER_RECTYPE, process order trailer line */
    else if (strcmp(record_type, TRAILER_RECTYPE) == 0)
    {
      /* read in order total and status */
      read_len = ORDER_TOTAL_LEN;
      num_bytes = read(file_desc, buff, read_len);
      buff[num_bytes] = '\0';

      /* if line is for the current order number, get order info */
      if (strcmp(file_ordrnum, orderno) == 0)
      {
        /* get order total from buff */
        field_len = ORDER_TOTAL_LEN - 1;
        field_start = 0;
        get_field(buff, c_order_total, field_start, field_len);

        /* get order status from buff */
        field_len = 1;
        field_start = ORDER_TOTAL_LEN - 1;
        get_field(buff, order_status, field_start, field_len);

        /* if order not pending, already signed */
        if (strcmp(order_status, "P") != 0)
        {
          status = TRANSD_ALREADY_SIGNED;
          goto Exit;
        }
        /* else specify order as signed */
        else
        {
          /* move file position back one character to write over P with S */
          lseek(file_desc, -1L, 1);

          if (print_char_field(file_desc, "S", 1) == 0)
          {
            logger(LOG_ERR, "%s", "Error writing signed status field to file.");
            status = TRANSD_ORDER_FILE_ERROR;
            goto Exit;
          }
        }

        /* set flag to true to specify found whole correct order */
        found_order = 1;
      }

      /* read in rest of order trailer line, the \n */
      read_len = 1;
      num_bytes = read(file_desc, buff, read_len);
      buff[num_bytes] = '\0';
    }
    /* if record type not valid, error */
    else
    {
      logger(LOG_ERR, "%s", "Invalid record type found in orders file.");
      status = TRANSD_ORDER_FILE_ERROR;
      goto Exit;
    }

    /* get next order number and record type from order file */
    read_len = ORDER_NUM_LEN + RECORD_TYPE_LEN - 2;
    num_bytes = read(file_desc, buff, read_len);
    buff[num_bytes] = '\0'; 
  } /* end of code to read in lines of orders file */

  /* check if complete order found */
  if (found_order == 0)
  {
    status = TRANSD_NOSUCH_ORDER;
    goto Exit;
  }

  /* split up expiration month and year */
  order_total = atof(c_order_total);
  if (strlen(ccexp) == 4)
  {
    ccexp_month[0] = ccexp[0];
    ccexp_month[1] = ccexp[1];
    ccexp_month[2] = '\0';
  }
  else
  {
    ccexp_month[0] = ccexp[0];
    ccexp_month[1] = '\0';
  }

  ccexp_year[0] = ccexp[strlen(ccexp) - 2];
  ccexp_year[1] = ccexp[strlen(ccexp) - 1];
  ccexp_year[2] = '\0';

  /* set up American Express information structure */
  strcpy(AMEXdata.acct_no,ccnum);
         AMEXdata.proc_code = CC_AUTHENTICATION;
         AMEXdata.amount = order_total;
  strcpy(AMEXdata.trace,orderno);
         AMEXdata.exp_month = atoi(ccexp_month);
         AMEXdata.exp_year = atoi(ccexp_year);
         AMEXdata.merchant_type = MERCHANT_TYPE;
  strcpy(AMEXdata.country,COUNTRY);
  strcpy(AMEXdata.pos_data,POS_DATA);
         AMEXdata.card_acceptor_bus_code = 5999;  /* was MERCHANT_TYPE */
         AMEXdata.approval_code_len = APPROVAL_CODE_LEN;
  strcpy(AMEXdata.card_accept_id,CARD_ACCEPTOR_ID);
  strcpy(AMEXdata.acceptor_loc.name,ACCEPTOR_NAME);
  strcpy(AMEXdata.acceptor_loc.zip,ACCEPTOR_ZIP);
  strcpy(AMEXdata.acceptor_loc.region,ACCEPTOR_REGION);
  strcpy(AMEXdata.acceptor_loc.country,ACCEPTOR_COUNTRY);
  strcpy(AMEXdata.currency_code, CURRENCY_CODE);
  strcpy(AMEXdata.security_info,SECURITY_INFO);
         AMEXdata.action_code = 1;
         AMEXdata.approval_code[0]='\0';
 
  /* perform American Express authorization */
  rc = amex_auth(&AMEXdata);

  /* update order file with action code, approval code, and status */
  if (update_order_file(&AMEXdata, rc) == BAD)
    logger(LOG_ERR, "%s", "Error updating orders file with American Express authorization/rejection info.");

  if (AMEXdata.action_code == APPROVED)
  {
    logger(LOG_ERR, "%s", "American Express authorization succeeded.");
    status = 0;
    goto Exit;
  }
  else
  {
    logger(LOG_ERR, "%s", "Error trying to get American Express authorization for charge.");
    status = TRANSD_AMEXAUTH_ERROR;
    goto Exit;
  }

  /* Exit gracefully */
  Exit:
  {
    if (open_flag == 1)
      unlock_close_file(file_desc, DBNAME);

    if (status == 0)
    {
      msg = (char*) malloc(strlen(errmsg[status]) + 128);
      sprintf(msg, "$%6.2f has been charged to %s's account. Return code = %s", order_total, user, errmsg[status]);
    }
    else
    {
      msg = (char*) malloc(strlen(errmsg[status]) + 1);
      strcpy(msg, errmsg[status]);
    }

    *pmsg = msg;

    return status;
  }
} /* OrderAccept */


/*---------------------------------------------------------------------------*/
/* update_order_file                                                         */
/*---------------------------------------------------------------------------*/
int update_order_file(struct AMEXOrd *AMEXdata, int AMEXrc)
{
  int  file_desc;
  int  updated_order = 0;   /* flag */
  int  num_bytes;     /* number of bytes to read from file */
  int  field_len;
  int  field_start;
  char file_ordrnum[ORDER_NUM_LEN];  
  char record_type[RECORD_TYPE_LEN];  
  int  read_len;
  int  write_len;
  char buff[ORDER_BUFF];
  char write_buff[ORDER_BUFF];
  char exp_date[CC_EXP_LEN];
  char act_code[ACTION_CODE_LEN];
  char status[2];

  /* open and lock the orders file */
  if ((file_desc = open_lock_file(DBNAME, O_RDWR)) == -1)
    return(BAD);

  /* rewind file before reading from it */
  lseek(file_desc, 0L, 0);

  /* read first order number and record type from orders file */
  read_len = ORDER_NUM_LEN + RECORD_TYPE_LEN - 2;
  num_bytes = read(file_desc, buff, read_len);
  buff[num_bytes] = '\0'; 

  /* for all lines in orders file until find whole order */
  while (num_bytes != 0 && updated_order == 0)
  {
    if (num_bytes == -1)
      continue;

    /* get order number from file buffer starting in col 0 of buff */
    field_len = ORDER_NUM_LEN - 1;
    field_start = 0;
    get_field(buff, file_ordrnum, field_start, field_len);

    /* get record type from buff */
    field_len = RECORD_TYPE_LEN - 1;
    field_start = ORDER_NUM_LEN - 1;
    get_field(buff, record_type, field_start, field_len);

    /* if record type = HEADER_RECTYPE, process order header line */
    if (strcmp(record_type, HEADER_RECTYPE) == 0)
    {
      /* if line is for the current order number, write credit card  info */
      if (strcmp(file_ordrnum, AMEXdata->trace) == 0)
      {
        /* read in header line up to CC info */
        read_len = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 4
                   + PHONE_LEN + ADDRESS_LEN * 4 + STATE_LEN + ZIP_LEN
                   + ACCOUNT_LEN * 3 + PERCENT_LEN * 3 - 20;
        num_bytes = read(file_desc, buff, read_len);
        buff[num_bytes] = '\0';

        /* print credit card info to file */
        if (print_numeric_field(file_desc, AMEXdata->acct_no, CC_NUM_LEN) == BAD)
          continue;

        sprintf(exp_date,"%d%d",AMEXdata->exp_month, AMEXdata->exp_year);
        if (print_numeric_field(file_desc, exp_date, CC_EXP_LEN) == BAD)
          continue;

        sprintf(act_code,"%d",AMEXdata->action_code);
        if (print_numeric_field(file_desc, act_code, ACTION_CODE_LEN) == BAD)
          continue;

        if (print_numeric_field(file_desc, AMEXdata->approval_code, APPROVAL_CODE_LEN+1) == BAD)
          continue;

        print_char_field(file_desc, "\n", 2);
      }
      else
      {
        /* read in rest of order header line */
        read_len = DATE_TIME_LEN + USERID_LEN + BOC_ACCT_LEN + NAME_LEN * 4
                   + PHONE_LEN + ADDRESS_LEN * 4 + STATE_LEN + ZIP_LEN
                   + ACCOUNT_LEN * 3 + PERCENT_LEN * 3 + CC_NUM_LEN
                   + CC_EXP_LEN + ACTION_CODE_LEN + APPROVAL_CODE_LEN - 22;
        num_bytes = read(file_desc, buff, read_len);
        buff[num_bytes] = '\0';
      }
    }
    /* if record type = COMMENT_RECTYPE, process order comment line */
    else if (strcmp(record_type, COMMENT_RECTYPE) == 0)
    {
      /* read in rest of order comments line */
      read_len = COMMENTS_LEN;
      num_bytes = read(file_desc, buff, read_len);
      buff[num_bytes] = '\0';
    }
    /* if record type = DETAIL_RECTYPE, process order detail line */
    else if (strcmp(record_type, DETAIL_RECTYPE) == 0)
    {
      /* read in rest of order detail line */
      read_len = ITEM_CODE_LEN + QUANTITY_LEN * 2 + PRICE_LEN + UNIT_LEN - 4;
      num_bytes = read(file_desc, buff, read_len);
      buff[num_bytes] = '\0';
    }
    /* if record type = TRAILER_RECTYPE, process order trailer line */
    else if (strcmp(record_type, TRAILER_RECTYPE) == 0)
    {
      /* if line is for the current order number, write status */
      if (strcmp(file_ordrnum, AMEXdata->trace) == 0)
      {
        /* read in order trailer line up to status field */
        read_len = ORDER_TOTAL_LEN - 1;
        num_bytes = read(file_desc, buff, read_len);
        buff[num_bytes] = '\0';

        /* determine status */
        if (AMEXrc == BAD)
          strcpy(status,"F");   /* authorization attempt failed */
        else
        {
          if ( AMEXdata->action_code == APPROVED)
            strcpy(status,"A");   /* approved */
          else
            strcpy(status,"J");  /* rejected */
        }

        /* write new status */
        if (print_char_field(file_desc, status, 1) == BAD)
          continue;
 
        /* set done reading specified order flag to true */
        updated_order = 1;
      }
      else
      {
        /* read in rest of order trailer line */
        read_len = ORDER_TOTAL_LEN + 1 ;
        num_bytes = read(file_desc, buff, read_len);
        buff[num_bytes] = '\0';
      }
    }
    /* if record type not valid, error */
    else
      continue;

    /* get next order number and record type from order file */
    read_len = ORDER_NUM_LEN + RECORD_TYPE_LEN - 2;
    num_bytes = read(file_desc, buff, read_len);
    buff[num_bytes] = '\0'; 
  } /* end of code to read in lines of orders file */

  unlock_close_file(file_desc, DBNAME);

  if (updated_order == 0)
    return BAD;
  else
    return GOOD;
} /* update_order_file */


/*---------------------------------------------------------------------------*/
/* open_lock_file: Open and lock file                                        */
/*---------------------------------------------------------------------------*/
int open_lock_file(char *file_name, int open_flag)
{
  int  file_desc;

  /* open file and create file descriptor */
  if ((file_desc = open(file_name, open_flag)) < 0)
  {
    logger(LOG_ERR, "%s %s %s", "Error opening the", file_name, "file.");
    return(file_desc);
  }

  /* rewind file before locking it */
  lseek(file_desc, 0L, 0);
  
  /* block until file unlocked, then lock */
  if (lockf(file_desc, F_LOCK, 0L) == -1)
  {
    logger(LOG_ERR, "%s %s %s", "Error locking the", file_name, "file.");
    close(file_desc);
    return(-1);
  }

  return(file_desc);
} /* open_lock_file */


/*---------------------------------------------------------------------------*/
/* get_field: Get next field from from buffer based on start and length      */
/*---------------------------------------------------------------------------*/
void get_field(char *buff, char *parm, int field_start, int field_len)
{
  int  buff_index, parm_index;

  /* adjust field_start and field_len to not use preceding blanks in string */
  while (buff[field_start] == ' ')
  {
    field_start++;
    field_len--;
  }

  /* set parm from buff */
  parm_index = 0;
  for (buff_index = field_start; buff_index < field_start + field_len; buff_index++)
  {
    parm[parm_index] = buff[buff_index];
    parm_index++;
  }
  parm[parm_index] = '\0';

  /* strip blanks off the end of the parm string */
  strip_blanks(parm);

  return;
} /* get_field */


/*---------------------------------------------------------------------------*/
/* strip_blanks: Strip blanks off the end of the argument string passed in   */
/*---------------------------------------------------------------------------*/
void strip_blanks(char *parm)
{
  int       index;
  size_t    parm_len;

  /* set parm_len with the length of the parm argument string */
  parm_len = strlen(parm);

  /* loop through chars in parm from right to left */
  for (index = parm_len - 1; index >= 0; index--)
  {
    /* if current char in parm is a space, set it to end of string char */
    if (isspace(parm[index]) != 0)
      parm[index] = '\0';
    /* at last non-blank character in parm, first from right, leave loop */
    else
      index = -1;
  }

  return;
} /* strip_blanks */


/*---------------------------------------------------------------------------*/
/* unlock_close_file: Unlock and close file                                  */
/*---------------------------------------------------------------------------*/
int unlock_close_file(int file_desc, char *file_name)
{
  /* rewind file before unlocking it */
  lseek(file_desc, 0L, 0);

  /* unlock file */
  if (lockf(file_desc, F_ULOCK, 0L) == -1)
  {
    logger(LOG_ERR, "%s %s %s", "Error unlocking the", file_name, "file.");
    return(0);
  }

  /* close file */
  close(file_desc);

  return(1);
} /* unlock_close_file */


/*---------------------------------------------------------------------------*/
/* print_numeric_field: Print numeric field to order file right justified    */
/*---------------------------------------------------------------------------*/
int print_numeric_field(int file_desc, char *field, int max_len)
{
  int  field_len, index;
  
  field_len = strlen(field);

  for (index = 1; index < max_len - field_len; index++)
  {
    /* write preceding blanks to file */
    if (write(file_desc, " ", 1) != 1)
    {
      logger(LOG_ERR, "%s %s %s", "Error writing to the", DBNAME, "file.");
      unlock_close_file(file_desc, DBNAME);
      return(0);
    }
  }

  /* write field to file */
  if (write(file_desc, field, field_len) != field_len)
  {
    logger(LOG_ERR, "%s %s %s", "Error writing to the", DBNAME, "file.");
    unlock_close_file(file_desc, DBNAME);
    return(0);
  }  

  return(1);
} /* print_numeric_field */


/*---------------------------------------------------------------------------*/
/* print_char_field: Print alphanumeric field to order file left justified   */
/*---------------------------------------------------------------------------*/
int print_char_field(int file_desc, char *field, int max_len)
{
  int  field_len, index;

  field_len = strlen(field);

  /* write field to file */
  if (write(file_desc, field, field_len) != field_len)
  {
    logger(LOG_ERR, "%s %s %s", "Error writing to the", DBNAME, "file.");
    unlock_close_file(file_desc, DBNAME);
    return(0);
  }

  for (index = 1; index < max_len - field_len; index++)
  {
    /* write succeding blanks to file */
    if (write(file_desc, " ", 1) != 1)
    {
      logger(LOG_ERR, "%s %s %s", "Error writing to the", DBNAME, "file.");
      unlock_close_file(file_desc, DBNAME);
      return(0);
    }
  }

  return(1);
} /* print_char_field */
